/************************************************************************
 * I N S T A L L _ S I L O                                              *
 ************************************************************************
 * Copyright (C) 2007 by                                                *
 * Dipl.-Ing. Michael Niederle                                          *
 *                                                                      *
 * This program is free software; you can redistribute it and/or modify *
 * it under the terms of the GNU General Public License as published by *
 * the Free Software Foundation; either version 2 of the License, or    *
 * (at your option) any later version.                                  *
 *                                                                      *
 * This program is distributed in the hope that it will be useful,      *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of       *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the         *
 * GNU General Public License for more details.                         *
 *                                                                      *
 * For details of the GNU General Public License see                    *
 * http://www.gnu.org/copyleft/gpl.html                                 *
 * or write to the                                                      *
 * Free Software Foundation, Inc.,                                      *
 * 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.            *
 ************************************************************************/

// this software is based upon the Linux installer program for SYSLINUX by
// H. Peter Anvin & Lars Munch Christensen
//
// Copyright 2003 Lars Munch Christensen - All Rights Reserved
// Copyright 1998-2007 H. Peter Anvin - All Rights Reserved

#include "mswin.h"
#include <winioctl.h>
#include <stdio.h>
#include <ctype.h>
#include "standard.h"

PROCEDURE Error(CHARACTER *msg);

#define SECTOR_SIZE 512U

CONSTANT UINT32 CHECKSUM = 0x369cf258;

ENUMERATION BOOT_SECTOR_OFFSETS IS
  bsJump            = 0x00,
  bsOemName         = 0x03,
  bsBytesPerSec     = 0x0b,
  bsSecPerClust     = 0x0d,
  bsResSectors      = 0x0e,
  bsFATs            = 0x10,
  bsRootDirEnts     = 0x11,
  bsSectors         = 0x13,
  bsMedia           = 0x15,
  bsFATsecs         = 0x16,
  bsSecPerTrack     = 0x18,
  bsHeads           = 0x1a,
  bsHiddenSecs      = 0x1c,
  bsHugeSectors     = 0x20,

  /* FAT12/16 only */
  bs16DriveNumber   = 0x24,
  bs16Reserved1     = 0x25,
  bs16BootSignature = 0x26,
  bs16VolumeID      = 0x27,
  bs16VolumeLabel   = 0x2b,
  bs16FileSysType   = 0x36,
  bs16Code          = 0x3e,

  /* FAT32 only */
  bs32FATSz32       = 36,
  bs32ExtFlags      = 40,
  bs32FSVer         = 42,
  bs32RootClus      = 44,
  bs32FSInfo        = 48,
  bs32BkBootSec     = 50,
  bs32Reserved      = 52,
  bs32DriveNumber   = 64,
  bs32Reserved1     = 65,
  bs32BootSignature = 66,
  bs32VolumeID      = 67,
  bs32VolumeLabel   = 71,
  bs32FileSysType   = 82,
  bs32Code          = 90,

  bsSignature     = 0x1fe
END;

#define bsHead      bsBytesPerSec
#define bsCode	    bs32Code	/* The common safe choice */

CONSTANT CHARACTER *program;	       // name of program
CONSTANT CHARACTER *boot_image_name; // name of boot image file
CONSTANT CHARACTER *drive;		       // drive to install to

UINT32 boot_image_size;

UINT8 boot_image_buf[63*SECTOR_SIZE];

#define get_8(p) (*(UINT8 *)(p))
#define get_16(p) (*(UINT16 *)(p))
#define get_32(p) (*(UINT32 *)(p))

#define set_16(p, v) (*(UINT16 *)(p)) = v;

/*
 * Check to see that what we got was indeed an MS-DOS boot sector/superblock;
 * Return NULL if OK and otherwise an Error message;
 */
FUNCTION CONSTANT CHARACTER *CheckBootSector(
  IN CONSTANT UINT8 *bs)
IS
  int sectorsize;
  UINT16 reserved_sectors;
  UINT32 sectors, fatsectors;
  INT64 dsectors;
  UINT32 clusters;
  int rootdirents, clustersize;
  CONSTANT UINT8 *boot_sec_buf = bs;

  /* Must be 0xF0 or 0xF8..0xFF */
  ON get_8(boot_sec_buf+bsMedia) != 0xF0 AND
     get_8(boot_sec_buf+bsMedia) < 0xF8
    GOTO invalid;

  sectorsize = get_16(boot_sec_buf+bsBytesPerSec);
  IF sectorsize == 512 THEN
    ; /* ok */
  ELIF sectorsize == 1024 OR sectorsize == 2048 OR sectorsize == 4096 THEN
    return "Only 512-byte sectors are supported";
  ELSE
    goto invalid;
  FI

  clustersize = get_8(boot_sec_buf+bsSecPerClust);
  ON clustersize == 0 OR (clustersize & (clustersize-1))
    GOTO invalid; // Must be nonzero and a power of 2

  sectors = get_16(boot_sec_buf+bsSectors);
  sectors = sectors ? sectors : get_32(boot_sec_buf+bsHugeSectors);

  reserved_sectors = get_16(boot_sec_buf+bsResSectors);
  dsectors = sectors - reserved_sectors;

  fatsectors = get_16(boot_sec_buf+bsFATsecs);
  fatsectors = fatsectors ? fatsectors : get_32(boot_sec_buf+bs32FATSz32);
  fatsectors *= get_8(boot_sec_buf+bsFATs);
  dsectors -= fatsectors;

  rootdirents = get_16(boot_sec_buf+bsRootDirEnts);
  dsectors -= (rootdirents+sectorsize/32-1)/sectorsize;

  ON dsectors < 0 OR fatsectors == 0 GOTO invalid;

  clusters = (UINT32)(dsectors/clustersize);

  IF clusters < 0xFFF5 THEN
    /* FAT12 or FAT16 */
    ON NOT get_16(boot_sec_buf+bsFATsecs) GOTO invalid;
    IF get_8(boot_sec_buf+bs16BootSignature) == 0x29 THEN
      IF NOT memcmp(boot_sec_buf+bs16FileSysType, "FAT12   ", 8) THEN
	ON clusters >= 0xFF5 RETURN "More than 4084 clusters but claims FAT12";
      ELIF NOT memcmp(boot_sec_buf+bs16FileSysType, "FAT16   ", 8) THEN
	ON clusters < 0xFF5 RETURN "less than 4084 clusters but claims FAT16";
      ELIF memcmp(boot_sec_buf+bs16FileSysType, "FAT     ", 8) ) {
	STATIC char fserr[] = "Filesystem type \"????????\" not supported";
	memcpy(fserr+17, boot_sec_buf+bs16FileSysType, 8);
	return fserr;
      FI
    FI
  ELIF clusters < 0x0FFFFFF5 THEN
    // moving the FileSysType and BootSignature was a
    // lovely stroke of M$ idiocy
    ON get_8(boot_sec_buf+bs32BootSignature) != 0x29 OR
	     memcmp(boot_sec_buf+bs32FileSysType, "FAT32   ", 8)
	    GOTO invalid;
  ELSE
    goto invalid;
  FI

  // The total space needed is boot image size + 1 sector.

  // Sector 1 is skipped because on FAT32 partitions it will likely
  // contain the file system info sector.

  // If this sector is overwritten the system will continue to work
  // but IO-performance might suffer.

  IF boot_image_size+SECTOR_SIZE > reserved_sectors*SECTOR_SIZE THEN
    fprintf(stderr, "Too few reserved sectors (%d)!\n", reserved_sectors);
    exit(1);
  FI

  return NULL;

  invalid:
  return "This doesn't look like a valid FAT filesystem";
END

/*
 * Windows Error function
 */
PROCEDURE Error(
  IN CHARACTER *msg)
IS
  VOID *lpMsgBuf;

  /* Format the Windows Error message */
  FormatMessage(
		FORMAT_MESSAGE_ALLOCATE_BUFFER |
		FORMAT_MESSAGE_FROM_SYSTEM |
		FORMAT_MESSAGE_IGNORE_INSERTS,
		NULL, GetLastError(),
		MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), // Default language
		(LPTSTR) &lpMsgBuf, 0, NULL);

  /* Print it */
  fprintf(stderr, "%s: %s", msg, (CHARACTER*) lpMsgBuf);

  /* Free the buffer */
  LocalFree(lpMsgBuf);
END

/*
 * Wrapper for ReadFile suitable for libfat
 */
/*INT32 libfat_readfile(intptr_t pp, void *buf, size_t secsize, libfat_sector_t sector)
{
  uint64_t offset = (uint64_t)sector * secsize;
  LONG loword = (LONG)offset;
  LONG hiword  = (LONG)(offset >> 32);
  LONG hiwordx = hiword;
  ULONG bytes_read;

  if (SetFilePointer((HANDLE)pp, loword, &hiwordx, FILE_BEGIN) != loword OR
       hiword != hiwordx OR
       !ReadFile((HANDLE)pp, buf, secsize, &bytes_read, NULL) OR
       bytes_read != secsize) {
    fprintf(stderr, "Cannot read sector %u\n", sector);
    exit(1);
  }

  return secsize;
}*/

TERMINATOR usage() IS
  fprintf(
    stderr,
    "Usage: install_silo <image-file> <drive-letter>:\n");
  exit(1);
END

FUNCTION INT32 main(
  IN INT32 argc,
  IN CHARACTER *argv[])
IS
  HANDLE boot_image_handle;
  HANDLE boot_drive_handle;
  ULONG bytes_read;
  ULONG bytes_written;

  UINT8 boot_sec_buf[512];
  CHARACTER drive_name[] = "\\\\.\\?:";
  CHARACTER drive_root[] = "?:\\";
  CONSTANT CHARACTER *errmsg;
  CONSTANT CHARACTER *bootsecfile = NULL;
  CONSTANT CHARACTER *subdir = NULL;

  IF (argc != 3) THEN
    usage();
  FI

  program = argv[0];
  boot_image_name = argv[1];
  drive = argv[2];

  boot_image_handle =
    CreateFile(
      boot_image_name, GENERIC_READ, FILE_SHARE_READ,
		  NULL, OPEN_EXISTING, 0, NULL);
  IF boot_image_handle == INVALID_HANDLE_VALUE THEN
    Error("Could not open file");
    exit(1);
  FI
		
  GetFileSizeEx(boot_image_handle, (PLARGE_INTEGER)&boot_image_size);
  IF boot_image_size < 1024 THEN
    fprintf(stderr, "Boot image file seems to be too small - less than 1 KB!\n");
    exit(1);
  ELIF boot_image_size > 62*SECTOR_SIZE THEN
    fprintf(stderr, "Boot image file is too large (max. 31 KB)!\n");
    exit(1);
  FI

  IF NOT drive OR NOT isalpha(drive[0]) OR drive[1] != ':' OR drive[2] THEN
    usage();
  FI

  /* Test if drive exists */
  ULONG drives = GetLogicalDrives();
  IF NOT(drives & (1 << (tolower(drive[0]) - 'a'))) THEN
    fprintf(stderr, "No drive %c:!\n", drive[0]);
    exit(1);
  FI

  /* Determines the drive type */
  drive_name[4] = drive[0];
  drive_root[0] = drive[0];
  UINT32 drive_type = GetDriveType(drive_root);

  /*
   * First open the drive
   */
  boot_drive_handle =
    CreateFile(
      (CONSTANT CHARACTER *)&drive_name, GENERIC_READ|GENERIC_WRITE,
		  FILE_SHARE_READ|FILE_SHARE_WRITE,
		  NULL, OPEN_EXISTING, 0, NULL);

  IF boot_drive_handle == INVALID_HANDLE_VALUE THEN
    Error("Could not open drive");
    exit(1);
  FI

  /*
   * Make sure we can read the boot sector
   */
  IF NOT ReadFile(boot_drive_handle, boot_sec_buf, 512, &bytes_read, NULL) THEN
    Error("Reading boot sector");
    exit(1);
  FI

  IF bytes_read != 512 THEN
    fprintf(stderr, "Could not read the whole boot sector!\n");
    exit(1);
  FI

  /* Check to see that what we got was indeed an MS-DOS boot sector/superblock */
  IF errmsg = CheckBootSector(boot_sec_buf) THEN
    fprintf(stderr, "%s!\n", errmsg);
    exit(1);
  FI

  IF NOT
    ReadFile(
      boot_image_handle, boot_image_buf, boot_image_size,
      &bytes_read, NULL)
  THEN
    Error("Reading boot image");
    exit(1);
  FI

  IF bytes_read != boot_image_size THEN
    fprintf(stderr, "Could not read the whole boot image!\n");
    exit(1);
  FI

  // pad last sector with zero-bytes
  WHILE boot_image_size&0x1ff DO
    boot_image_buf[boot_image_size++] = 0;
  END

  // copy disk parameter block
  memcpy(boot_image_buf+bsHead, boot_sec_buf+bsHead, bsCode-bsHead);

  // patch number of remaining sectors of boot loader
  UINT16 remaining_sectors = (UINT16)(boot_image_size/SECTOR_SIZE)-1;
  set_16(boot_image_buf+0x1fa, remaining_sectors);

  // compute checksum
  UINT32 checksum = 0;
  UINT32 *p = (UINT32 *)(boot_image_buf+0x200);
  ForTimes(remaining_sectors*(SECTOR_SIZE/sizeof(UINT32))) DO
    checksum += *p++;
  NEXT

  // patch checksum value
  *(UINT32 *)(boot_image_buf+0x204) = CHECKSUM-checksum;

  /* write boot loader image */
  LARGE_INTEGER position;
  position.QuadPart = 0;
  UINT64 new_position;
  IF NOT
    SetFilePointerEx(
      boot_drive_handle, position, (PLARGE_INTEGER)&new_position, FILE_BEGIN)
  THEN
    Error("Could not move file pointer to start of boot drive");
    exit(1);
  FI

  IF new_position != 0 THEN
    fprintf(stderr, "File pointer movement to start of boot drive failed!\n");
    exit(1);
  FI

  IF NOT
    WriteFile(
      boot_drive_handle, boot_image_buf, SECTOR_SIZE,
      &bytes_written, NULL)
  THEN
    Error("Could not write boot sector");
    exit(1);
  FI

  IF bytes_written != SECTOR_SIZE THEN
    fprintf(stderr, "Could not write whole boot sector!\n");
    exit(1);
  FI

  position.QuadPart = 2*SECTOR_SIZE;
  IF NOT
    SetFilePointerEx(
      boot_drive_handle, position, (PLARGE_INTEGER)&new_position, FILE_BEGIN)
  THEN
    Error("Could not move file pointer to the start of sector 2");
    exit(1);
  FI

  IF NOT
    WriteFile(
      boot_drive_handle, boot_image_buf+SECTOR_SIZE, boot_image_size-SECTOR_SIZE,
      &bytes_written, NULL)
  THEN
    Error("Could not write boot image");
    exit(1);
  FI

  IF bytes_written != boot_image_size-SECTOR_SIZE THEN
    fprintf(stderr, "Could not write whole boot image!\n");
    exit(1);
  FI

  /* flush the media */
  IF NOT FlushFileBuffers(boot_drive_handle) THEN
    Error("FlushFileBuffers failed");
    exit(1);
  FI

  CloseHandle(boot_drive_handle);

  /* Done! */
  printf("Bootloader successfully installed onto drive %s\n", drive);
  return 0;
END
